-- Time-stamp: <Fri Jan 23 1998 21:05:34 Stardate: [-30]0599.39 hwloidl>
-- Barnes-Hut.mir converted to Haskell by hwloidl on Thu Dec 5 01:38:04 GMT 1996
-- 
-- N-Body Particle Simulation Kernel
--
-- Author:  Tony Field
-- Date:    November 1996
--
-- Parallel Haskell version: Hans-Wolfgang Loidl <hwloidl@dcs.glasgow.ac.uk>
--
-- Intro:   This is a very simple implementation of the main iteration
--          function in the Barnes-Hut particle simulation algorithm.
--          It is not intended to be used for particle simulation, but
--          rather to see how easy it might be to specify in a functional
--          language.  It also serves as a useful reference source for
--          students.
--
-- Warning: None of this ord is guaranteed, although it is correctly typed.
--
-------------------------------------------------------------------------------
--
-- Particle positions, velocities and resultant forces are vectors, 
-- represented as lists for simplicity.  
-- The ord works for both two and three dimensions with only one small
-- change required (to zerovec) to switch between the two.
-- For 2D vectors are of the form [x,y] and the algorithm
-- works on quad-trees; for 3D vectors are of the form [x,y,z] and the
-- ord will automatically produce oct-trees.
--  
-------------------------------------------------------------------------------

--@node Barnes-Hut Algorithm
--@chapter Barnes-Hut Algorithm, ,

module BarnesHut where

-----------------------------------------------------------------------------
--@menu
--* Imports::			
--* Types::			
--* Aux fcts::			
--* Problem specific::		
--* Barnes-Hut Main::		
--* Index::			
--@end menu

--@node Imports, Types, Barnes-Hut Algorithm, Barnes-Hut Algorithm
--@section Imports
-----------------------------------------------------------------------------

-- import Mira2hs

import Strategies
import List (transpose)

-----------------------------------------------------------------------------
--@node Types, Aux fcts, Imports, Barnes-Hut Algorithm
--@section Types
-----------------------------------------------------------------------------

--@menu
--* Basic Types::		
--* Trees::			
--* Fcts on Trees::		
--@end menu

--@node Basic Types, Trees, Types, Types
--@subsection Basic Types

type Position     = [Float]

type Velocity     = [Float]

type Forcevec     = [Float]

type Particlemass = Float

--
-- We begin with a "stock" tree data type and some useful h.o.f.s
--
--@node Trees, Fcts on Trees, Basic Types, Types
--@subsection Trees

--@cindex parleaves
--@cindex tmap
--@cindex treduce

data Tree a b c = Empty a | Leaf b | Node c [Tree a b c]

--type QuadTree a = Tree Bbox (Bbox,Particle) a 
      -- represents space


instance (NFData a, NFData b, NFData c) => NFData (Tree a b c) where
  rnf (Empty x) = rnf x
  rnf (Leaf x) = rnf x
  rnf (Node x l) = rnf x `seq` rnf l

-- Reduce to spine normal form (i.e. structure of tree but ignore values)
rspinenf :: (NFData a, NFData b, NFData c) =>  Strategy (Tree a b c)
rspinenf  (Empty _) = ()
rspinenf  (Leaf _) = ()
rspinenf  (Node _ l) = parList rspinenf l `seq` ()

-- Reduce all leaves in the tree in parallel
parleaves :: (NFData a, NFData b, NFData c) => 
	     Strategy b -> Strategy (Tree a b c)
parleaves s (Empty _) = ()
parleaves s (Leaf x) = s x
parleaves s (Node x l) = parList (parleaves s) l 


tmap :: (NFData a, NFData b, NFData c, NFData a', NFData b', NFData c') =>
        (a -> a') -> (b -> b') -> (c -> c') -> Tree a b c -> Tree a' b' c'
tmap f1 f2 f3 ( Empty v )
  = Empty ( f1 v )
tmap f1 f2 f3 ( Leaf v )
  = Leaf ( f2 v )
tmap f1 f2 f3 ( Node v ts )
  = Node node list 

    `using` \ _ -> rnf node `par`
                   parList rnf list 

    where node = f3 v
          list = map ( tmap f1 f2 f3 ) ts

-- ngo' ngoq:  = Node ( f3 v ) ( map ( tmap f1 f2 f3 ) ts )

treduce :: (NFData a, NFData b, NFData c, NFData d) =>
        (a -> d) -> (b -> d) -> (c -> [d] -> d) -> Tree a b c -> d
treduce f1 f2 f3 ( Empty v )
  = f1 v
treduce f1 f2 f3 ( Leaf v )
  = f2 v
treduce f1 f2 f3 ( Node v ts )
  = f3 v list 

    `using` \ _ -> parList rnf list 

    where list = map ( treduce f1 f2 f3 ) ts

--
-- We will use three specialisations of the general tree type.  The 
-- 'buildtree' phase will insert the particles into a global tree formed
-- by recursive decomposition of the space domain.  Each constructor will
-- contain a bounding box (the bounding volume for the subdomain rooted
-- at the constructor).  Leaf nodes contain a single particle in addition.
-- In computing the mass centres for each subdomain a new tree is built in
-- which the Node constructor holds the centre of mass for the subtrees beneath
-- it.  The centre of mass has the same type as a particle to avoid type
-- coercions later on.  In computing the net forces on each particle, 
-- another tree is built in which the particles at the leaves are replaced
-- by (particle,forcevector) pairs, where the forcevector is the resultant
-- force acting on the particle by other particles and aggregated subdomains.
--
-- In production C equivalents there is one tree with initially empty fields
-- that are filled in by the various phases.
--
-- A bounding box is a pair of vectors (it is not essential to store these
-- with the constructors as they can be computed on the fly) and a particle
-- comprises its position, velocity vector and mass.
-- 
-- We define some useful h.o.f.s for the specific trees in question.  The
-- specialised tree can be re-used in other adaptive codes and to implement
-- new tree operations.
--

type Bbox     = ( Position, Position )

type Particle = ( Position, Velocity, Particlemass )

--
-- mapleaf applies a given function to each leaf value
--
--@node Fcts on Trees,  , Trees, Types
--@subsection Fcts on Trees

--@cindex mapleaf
--@cindex partreduce
--@cindex upsweep
--@cindex flatten

mapleaf :: (NFData a, NFData b, NFData c, NFData d, NFData e) =>
	   (b -> d) -> (Tree a (e,b) c) -> (Tree a (e,d) c)
mapleaf f 
  = tmap id f' id
    where
      f' ( bb, v ) = ( bb, f v )

--
-- partreduce reduces the tree selectively according to a "long-range" 
-- interaction criteria.  If an internal node satisfies a given predicate
-- ('farfrom') a node interaction is applied;  if not the interactions
-- are applied to the subtrees and the results combined using 'combine'.
-- Base cases are provided for Empty trees ('basecase') and 
-- leaves ('leafinteraction').
--
partreduce basecase leafinteraction farfrom nodeinteraction combine t
  = treduce f1 f2 f3 t 
    where
      f1 bb         = basecase
      f2 (bb,p')    = leafinteraction p' 
      f3 (bb,cm) ts | (farfrom bb) = nodeinteraction cm
                    | otherwise = combine ts

--
-- upsweep applies one function to the values at the leaves of the tree
-- and another at the nodes.  The function applied to the nodes combines 
-- the results of the applications to the subtrees.  So at each step we
-- deliver the newly constructed tree and the result of the upsweep 
-- for use higher up. 
--
upsweep lfun nfun t
  = snd ( treduce f1 f2 f3 t )
    where
      f1 bb     = ( nullparticle, Empty bb )
      f2 (bb,p) = ( v, Leaf (bb,v) )
                  where
                    v = lfun p
      f3 bb ts  = ( v, Node (bb,v) ( map snd ts ) )
                  where
                    v = nfun ( map fst ts )

--
-- The inevitable flatten...
--
flatten :: (NFData a, NFData b, NFData c, NFData e) => 
	   (Tree a (e,b) c) -> [b]
flatten 
  = treduce f1 f2 f3
    where
      f1 v      = []
      f2 (bb,p) = [p]
      f3 v ts   = concat ts

--
-- Some useful functions not defined in the prelude
--
--@node Aux fcts, Problem specific, Types, Barnes-Hut Algorithm
--@section Aux fcts

--@cindex map3
--@cindex mkpair
--@cindex zerovec
--@cindex vsum
--@cindex square

map3 = zipWith3

--map3 :: (a -> b -> c -> d) -> [a] -> [b] -> [c] -> [d]
--map3 f [] b c 
--  = []
--map3 f (a:as) (b:bs) (c:cs) 
--  = f a b c : map3 f as bs cs

mkpair :: (a -> b) -> (a -> c) -> a -> (b, c)
mkpair f g x
  = ( f x, g x )

zerovec :: Forcevec
zerovec = [0,0]    -- Note:  in 3D this is [0,0,0]

vsum :: [Forcevec] -> Forcevec
vsum = foldr ( zipWith (+) ) zerovec

square x
  = x*x

--@node Problem specific, Barnes-Hut Main, Aux fcts, Barnes-Hut Algorithm
--@section Problem specific
-- 
-- Problem specific stuff...
--

--@cindex g
--@cindex nullparticle
--@cindex posn
--@cindex vel
--@cindex mass

g :: Particlemass
g = 10 

nullparticle :: Particle
nullparticle
  = ( zerovec, zerovec, 0 )

posn :: Particle -> Position
posn ( p, v, m )
  = p 

vel :: Particle -> Velocity
vel ( p, v, m )
  = v

mass :: Particle -> Particlemass
mass (p, v, m )
  = m 

------------------------------------------------------------------------------

--@node Barnes-Hut Main, Index, Problem specific, Barnes-Hut Algorithm
--@section Barnes-Hut Main

--
-- Barnes-Hut begins here...
--

--
-- The functions to support 'buildtree'...
--
-- locate determines to which subtree a particle belongs; the subtree
-- bounding box is supplied
--
-- buildempty constructs a new node with all empty subtrees.  This is used
-- when a particle is inserted into an empty tree
--
-- buildboxes constructs the bounding boxes for each subtree
--
-- boundbox constructs the bounding box for a given set of particles
--
-- update adds a particle to its appropriate subtree
--
-- insert adds a particle to a tree
--
-- buildtree adds a collection (list) of particle to an empty tree
--

--@menu
--* Basic Fcts::		
--* Compute Centres of Mass::	
--* Force Calculation::		
--* Particle Update::		
--* Main Iteration Fct::	
--@end menu

--@node Basic Fcts, Compute Centres of Mass, Barnes-Hut Main, Barnes-Hut Main
--@subsection Basic Fcts

--@cindex locate
--@cindex buildempty
--@cindex buildboxes
--@cindex boundbox
--@cindex insert
--@cindex buildtree

locate :: Position -> Bbox -> Int
locate p (p1, p2) 
  = sum ( zipWith (*) [1,2,4] ( map3 f p p1 p2 ) )
    where
      f a b c | (a > b+(c-b)/2) = 1
              | (otherwise ) = 0

buildempty :: Bbox -> Tree Bbox (Bbox, Particle) Bbox
buildempty (p1,p2)
  = Node (p1,p2) ( map Empty ( buildboxes p1 p2 midvec ) )
    where
      midvec = zipWith middle p1 p2
      middle s t = s + (t-s)/2

buildboxes [] [] []             
  = [([],[])]
buildboxes (x:xs) (y:ys) (z:zs) 
  = map addx l ++ map addx' l
    where
      l = buildboxes xs ys zs 
      addx  (a,b) = (x:a, z:b)
      addx' (a,b) = (z:a, y:b)

boundbox ps
  = ( map minimum ords, map maximum ords )
    where
      ords = transpose ps

update :: [a] -> Int -> a -> [a]  
update (x:xs) i v
  | (i==0) = v : xs
  | otherwise = x : update xs (i-1) v

insert :: Particle -> Tree Bbox (Bbox,Particle) Bbox -> Tree Bbox (Bbox,Particle) Bbox
insert p ( Empty bb ) 
  = Leaf (bb,p)
insert p ( Leaf (bb,p') ) 
  = insert p ( insert p' ( buildempty bb ) ) 
insert p ( Node bb ts )
  = Node bb ( update ts i t )
    where 
      i = locate ( posn p ) bb
      t = insert p ( ts!!i ) 


par_foldr :: (NFData a, NFData b) => (a -> b -> b) -> b -> [a] -> b
par_foldr f z [] = z
par_foldr f z (x:xs) = res
	               `using` \ _ -> rwhnf x' `par`
		                      rnf res
	               where x' = par_foldr f z xs
			     res = x `f` x'


buildtree :: [Particle] -> Tree Bbox (Bbox, Particle) Bbox
buildtree ps
  = foldr {-par_foldr-} insert ( Empty ( boundbox ( map posn ps ) ) ) ps 


    
--@node Compute Centres of Mass, Force Calculation, Basic Fcts, Barnes-Hut Main
--@subsection Compute Centres of Mass
--
-- The ord to compute the centres of mass is trivial using the h.o.f.s 
-- provided
--

--@cindex centremass
--@cindex computemasscentres 

-- Applied to each subtree of a Node in a QuadTree
centremass :: [Particle] -> Particle
centremass ps 
  = ( cm, zerovec, totalmass )
    where
      cm = vsum ( map scale ps )
      scale ( p, v, m ) = map ( (*) ( m/totalmass ) ) p
      totalmass = sum ( map mass ps )

computemasscentres :: Tree Bbox (Bbox, Particle) Bbox -> 
                      Tree Bbox (Bbox, Particle) (Bbox, Particle)
computemasscentres ps
  = (upsweep id centremass) $ ps


--@node Force Calculation, Particle Update, Compute Centres of Mass, Barnes-Hut Main
--@subsection Force Calculation

--
-- Force calculation ord...
--
-- isfaraway is the opening criteria.  Use your favourite error-bounding
-- term in here.  When always false is equivalent to the O(n^2) algorithm.
--
-- forcebetween computes the force vector between two particles
--
-- computeforce uses partreduce to selectively reduce the global tree
-- using the opening criteria for long-range interactions.  Note that the
-- resultant force from the subtrees is the sum of the force vectors for
-- each.
--
-- Note that the main map pairs up each particle at the leaves with the
-- resultant force acting on that particle.
--

--@cindex isfaraway
--@cindex forcebetween
--@cindex computeforce
--@cindex computeforces

isfaraway a b
  = False 

forcebetween :: Particle -> Particle -> Forcevec
forcebetween (p,v,m) (p',v',m')
 | p==p' && v==v' && m==m'
  = zerovec
 | otherwise
  = map ( (*) (f/r) ) deltapos
    where
      deltapos = zipWith (-) p p'
      r = sqrt ( sum ( map square deltapos ) )
      f = g * m * m' / ( square r )

computeforce :: Tree Bbox (Bbox, Particle) (Bbox, Particle) -> 
                Particle -> 
		Forcevec
computeforce t p
  = partreduce zerovec pforce ( isfaraway p ) pforce vsum t
    where
      pforce = forcebetween p

computeforces :: Tree Bbox (Bbox, Particle) (Bbox, Particle) -> 
                 Tree Bbox (Bbox, (Particle, Forcevec)) (Bbox, Particle)
computeforces t
  = mapleaf ( mkpair id ( computeforce t ) ) t
    `using`  parleaves rnf

--@node Particle Update, Main Iteration Fct, Force Calculation, Barnes-Hut Main
--@subsection Particle Update
--
-- Particle update is trivial...
--
--@cindex updateparticle
--@cindex updatepos

-- Now, this looks quite dubious to me
updateparticle :: (Particle, Forcevec) -> Particle
updateparticle ( ( p, v, m ), f )
  = (p, v, m ) 

updatepos :: Tree Bbox (Bbox, (Particle, Forcevec)) (Bbox, Particle) -> 
             [Particle]
updatepos x = (flatten .|| parleaves rnf $  mapleaf updateparticle) $ x


--@node Main Iteration Fct,  , Particle Update, Barnes-Hut Main
--@subsection Main Iteration Fct
--
-- And finally the main iteration function
--
--@cindex iteration

iteration :: [Particle] -> [Particle]
{-
   This pipeline is strict with hardly any parallelism in it.
   If we compose it via `.|| rnf' we just create a huge number of blocked
   threads waiting for input from earlier stages.
-}
iteration = updatepos             .|| rwhnf $
	    computeforces         .|| rwhnf $
	    computemasscentres    .|| rwhnf $
	    buildtree

iterations :: Int -> [Particle] -> [Particle]
iterations 0 x = x
iterations n x = iterations (n-1) (iteration x)
